from __future__ import division
import Tkinter as tk
import tkMessageBox
import numpy as np


__version__ = 'v1.2'


# STANDARD VALUE DECADE FOR 1% RESISTORS
R_LOOKUP = np.array([10.0, 10.2, 10.5, 10.7, 11.0, 11.3, 11.5, 11.8, 12.1, 12.4, 12.7, 13.0, 13.3, 13.7, 14.0, 14.3, 
                     14.7, 15.0, 15.4, 15.8, 16.2, 16.5, 16.9, 17.4, 17.8, 18.2, 18.7, 19.1, 19.6, 20.0, 20.5, 21.0, 
                     21.5, 22.1, 22.6, 23.2, 23.7, 24.3, 24.9, 25.5, 26.1, 26.7, 27.4, 28.0, 28.7, 29.4, 30.1, 30.9, 
                     31.6, 32.4, 33.2, 34.0, 34.8, 35.7, 36.5, 37.4, 38.3, 39.2, 40.2, 41.2, 42.2, 43.2, 44.2, 45.2, 
                     46.4, 47.5, 48.7, 49.9, 51.1, 52.3, 53.6, 54.9, 56.2, 57.6, 59.0, 60.4, 61.9, 63.4, 64.9, 66.5, 
                     68.1, 69.8, 71.5, 73.2, 75.0, 76.8, 78.7, 80.6, 82.5, 84.5, 86.6, 88.7, 90.9, 93.1, 95.3, 97.6])

# STANDARD CAPACITOR VALUES (uFarads)                     
# https://www.petervis.com/electronics/Electrolytic_Capacitor_Values/Electrolytic_Capacitor_Values.html
C_LOOKUP = np.array([0.1, 0.15, 0.22, 0.33, 0.47, 0.68, 1, 1.5, 2, 2.2, 3, 3.3, 4, 4.7, 5, 5.6, 6.8, 7, 8, 8.2, 10, 
                     12, 15, 16, 18, 20, 21, 22, 24, 25, 27, 30, 33, 35, 36, 39, 40, 43, 47, 50, 53, 56, 60, 68, 72, 
                     75, 82, 88, 100, 108, 120, 124, 130, 140, 145, 150, 161, 170, 180, 189, 200, 210, 216, 220, 230, 
                     233, 240, 243, 250, 270, 300, 320, 324, 330, 340, 350, 370, 378, 380, 390, 400, 420, 430, 450, 460, 
                     470, 480, 500, 510, 520, 540, 550, 560, 590, 620, 645, 650, 680, 700, 708, 730, 800, 820, 850, 860, 
                     1000, 1100, 1200, 1300, 1400, 1500, 1600, 1700, 1800, 2000, 2100, 2200, 2500, 2600, 2700, 2800, 2900, 
                     3000, 3100, 3300, 3400, 3500, 3600, 3700, 3900, 4000, 4100, 4200, 4300, 4600, 4700, 4800, 5000, 5100, 
                     5400, 5500, 5600, 5800, 6000, 6500, 6800, 7200, 7400, 7600, 7800, 8200, 8300, 8400, 8700, 9000, 9600, 
                     10000, 11000, 12000, 13000, 15000, 16000, 17000, 18000, 20000, 22000, 23000, 24000, 25000, 26000, 
                     27000, 28000, 30000, 31000, 32000, 33000, 34000, 36000, 37000, 38000, 39000, 40000, 41000, 47000, 
                     48000, 50000, 55000, 56000, 60000, 62000, 66000, 68000, 76000])
#  #  #  #  #  #  #  #  #  #  #  #  #  #  #  #  #  #  #  #  #  #  #  #   


def gain_reduction(S, X):
    '''
    Inputs:
    S: geophone sensitivity (V/m/s)
    X: max. motion before clipping (mm/s)
    
    Outputs:
    G: gain reduction ratio 
    '''
    voltage_span = 2.048 # V
    G = voltage_span / (S * (1e-3 * X))
    return G
    
    
def max_reading(S, X, scale=1.0):
    '''
    Inputs:
    S: geophone sensitivity (V/m/s)
    X: max. motion before clipping (mm/s)
    scale: scaling factor (default 1)
    
    Outputs:
    M: maximum digitizer reading
    '''
    M = scale / gain_reduction(S, X)
    return M


def theoretical_solutions(data):
    '''
    Inputs:
    data: a dict of input values with the following fields expected
        S  : geophone sensitivity (V/m/s)
        X  : max. motion before clipping (mm/s)
        Rs : geophone coil resistance (Ohms)
        Rd : Jam specific value (Ohms)
        Rdr: geophone CDR (Ohms)
        D1 : current damping factor
        D2 : desired damping factor
        Cb : resistor in place of C[b] (MOhms)
        F  : low pass filter frequency (hz)
    
    Outputs:
    payload: a dict of calculated results and intermediate values
        with the following fields
        Ra    : (Ohms)
        Rb    : (Ohms)
        Rc    : Damping Resistance (Ohms)
        Rd    : (Ohms)
        Rm    : Effective Damping Resistance (Ohms)
        Rx    : Differential Rin (Ohms)
        G     : gain reduction ratio 
        Ca    : (nFarads)
        Cb    : (MOhms)
        omegaC: (rad / s * Farads)
        F     : (Hz)
    '''
    Ra = 0.5 * data['Rs']
    Rd = data['Rd']
    Rm = data['Rdr'] * data['D1'] / data['D2']    # EDR
    G = gain_reduction(data['S'], data['X']) # RR
    Rx = (2 * Rm * Rd / G) / (Rm + 2*Ra)
    Rc = (2 * Rm * Rd / G) / (2 * Rd / G - (Rm + 2*Ra))
    Rb = Rx / 2 - Rd
    F = data['F']
    Cb = data['Cb']
    omega = 2 * np.pi * F
    
    omegaC = (Rc*(Ra + Rb + Rd) + Ra*Rx) / (Rd * (Ra*Rc + Rb*Rc + 2*Ra*Rb))
    Ca = omegaC / omega * 1e9  # convert to nanoFarads
    
    payload = {
        'Ra': Ra, 'Rd': Rd, 'Rm': Rm, 'G': G, 
        'Rx': Rx, 'Rc': Rc, 'Rb': Rb,
        'Ca': Ca, 'Cb': Cb, 
        'omegaC': omegaC, 'F': F
    }
    return payload
    
    
def check_solutions(data):
    '''
    Inputs:
    data: a dict with the solution values 
          (return value of `theoretical_solutions` function)
          
    Outputs:
    flag: bool (True if solutions are valid)
    '''
    flag = data['Rb'] >= 0.0 and data['Rc'] >= 0.0 and data['G'] <= 1.0
    return flag
    
    
def generate_lookup(R): 
    '''
    Inputs:
    R: theoretical resistance value (Ohms)
    
    Outputs:
    target: a numpy array of real-world resistor
            values from the lookup smaller than 110% of R
            (we add 110% threshold, so a bit larger values 
            than R can be selected as well)
    '''
    p = int(np.log10(R) - 1)   # order of magnitude
    target = np.hstack([R_LOOKUP * 10**p, R_LOOKUP * 10**(p - 1)])
    # drop larger than R values (allowing 10% over)
    return target[target <= 1.1 * R]  


def select_component_values(data):
    '''
    Inputs: 
    data: a dict with the solution values 
          (return value of `theoretical_solutions` function)
          
    Outputs:
    payload: a dict of selected component values with the following fields
        Rb : theoretical value (Ohms)
        Rc : theoretical value (Ohms)
        Ca : theoretical value (Ohms)
        _Ra: selected value (Ohms)
        _Rb: selected value (Ohms)
        _Rc: selected value (Ohms)
        _Ca: selected value (nfarads)
        _Cb: same as theoretical value (MOhms)
        
    Note: values prefixed with an underscore are computed for
          real world components from the lookup table
    '''
    # compute Ca
    table = C_LOOKUP * 1e3 # uFarads to nFarads
    idx = np.argmin(np.abs(data['Ca'] - table))
    _Ca = table[idx]
    
    # Compute R[c]
    table = generate_lookup(data['Rc'])
    idx = np.argmin(np.abs(data['Rc'] - table))
    _Rc = table[idx]
    
    # Compute R[a] and R[b]
    table = generate_lookup(data['Rb'])
    # as per https://stackoverflow.com/a/11146645/3023116
    pairs = np.dstack(np.meshgrid(table, table)).reshape(-1, 2)
    # drop duplicates (a, b == b, a)
    pairs = pairs[pairs[:, 0] <= pairs[:, 1]]
    arrRb = pairs.sum(axis=1)
    
    derr = compute_errors(data, arrRb, _Rc, _Ca)
    # minimize total error
    error = np.hypot(derr['_Rm_tol'], derr['_G_tol'])
    idx = np.argmin(error)

    _Ra, _Rb = pairs[idx]
    
    payload = {
        'Rb': data['Rb'], 'Rc': data['Rc'], 'Ca': data['Ca'],
        '_Ra': _Ra, '_Rb': _Rb, '_Rc': _Rc, '_Ca': _Ca, '_Cb': data['Cb']
    }
    return payload
    
    
def compute_errors(data, _Rb, _Rc, _Ca):
    '''
    Inputs: 
    data: a dict with the solution values 
          (return value of `theoretical_solutions` function)
    _Rb : R[a] + R[b] (Ohms)
    _Rc : R[c] (Ohms)
    _Ca : C[a] (nfarads)
    
    Outputs:
    payload: a dict with theoretical, actual and error values
     
    Note: values prefixed with an underscore are computed for
          real world components from the lookup table
    '''
    _Rx = 2 * (data['Rd'] + _Rb)
    _Rm = _Rx * _Rc / (_Rx + _Rc)
    _Rm_tol = np.abs(data['Rm'] - _Rm) / data['Rm']

    _G = 2 * _Rc * data['Rd'] / (2 * data['Ra'] * (_Rx + _Rc) + _Rx * _Rc)
    _G_tol = np.abs(data['G'] - _G) / data['G']
    
    _F = data['omegaC'] / (2 * np.pi * _Ca * 1e-9)
    _F_tol = np.abs(data['F'] - _F) / data['F']
    
    payload = {
        'Rm': data['Rm'], '_Rm': _Rm, '_Rm_tol': _Rm_tol, 
        'G': data['G'], '_G': _G, '_G_tol': _G_tol,
        'F': data['F'], '_F': _F, '_F_tol': _F_tol,
    }
    return payload


#  #  #  #  #  #  #  #  #  #  #  #  #  #  #  #  #  #  #  #  #  #  #  #   


class Application(object):

    def __init__(self):
    
        self.small_font = "Sans 8"
        self.medium_font = "Sans 9"
        self.default_font = "Sans 11"
        self.table_title_font = "Sans 11 italic"
        self.frame_title_font = "Sans 12 bold"
        
        self.default_inputs = {
            'X' : 26, 
            'S' : 629,
	   'Rs' : 3600,
           'Rdr': 6525,
           'D1' : 1, 
           'D2' : 0.7, 
            'F' : 50, 
           'Rd' : 5100,   # RJAM-specific. Does not change
           'Cb' : 1,      # MOhms. Does not change
        }
    
        self.root = self.setup_main_window()
        
        self.inputs = self.setup_inputs()
        self.outputs = self.setup_outputs()
        self.multi_units = self.setup_multi_units()
        
        self.have_all_inputs = False

        # set default values for input variables
        for key, value in self.default_inputs.items():
            self.inputs[key]['var'].set(value)
        
        self.setup_components(self.root)
        # CLICK BUTTON
	self.calculate_callback(True)

        
    def setup_main_window(self, root=None):
    
        if root == None:
            root = tk.Tk()
        root.title('Determining Resistor Values for Connecting a Geophone')
        root.config(background="#aaccbb")
        root.option_add("*Font", self.default_font)
        return root
        
        
    def setup_multi_units(self):
        # units is a dropdown rather than a label
        units_map = {
            'X': ["mm/s", "in/s"],
            'S': ["V/m/s", "V/mm/s", "V/inch/s"]
        }
        converter_map = {
            'X': [lambda x: x, lambda x: 25.4 * x],
            'S': [lambda x: x, lambda x: 1000 * x, lambda x: x / 0.0254]
        }
        return {
            key: (tk.StringVar(value=units[0]), units, dict(zip(units, converter_map[key]))) 
            for key, units in units_map.items()
        }
                
                
    def setup_inputs(self):
        payload = {
            'X'  : {'label': 'Max. motion before clipping*', 'var': tk.DoubleVar(), 'units': "mm/s"},
            'S'  : {'label': 'Geophone Sensitivity'        , 'var': tk.DoubleVar(), 'units': "V/m/s"},
            'Rs' : {'label': 'Geophone coil Resistance'    , 'var': tk.IntVar()   , 'units': "Ohms"},
            'Rdr': {'label': 'Geophone CDR'                , 'var': tk.IntVar()   , 'units': "Ohms"},
            'D1' : {'label': 'For a damping factor of'     , 'var': tk.DoubleVar(), 'units': ''},
            'D2' : {'label': 'Desired damping factor'      , 'var': tk.DoubleVar(), 'units': ''},
            'Rd' : {'label': 'Jam Specific'                , 'var': tk.IntVar()   , 'units': "Ohms"},
            'F'  : {'label': 'Low Pass Filter Frequency**' , 'var': tk.IntVar()   , 'units': "Hz"},
            'Cb' : {'label': 'Resistor in place of C[b]'   , 'var': tk.IntVar()   , 'units': "MOhms"},
        }
        return payload
        
    
    def setup_outputs(self):
        payload = {
            'Rb'  : {'label': 'R[a] + R[b]', 'var': tk.IntVar()   , 'units': "Ohms"},
            'Rc'  : {'label': 'R[damp]'    , 'var': tk.IntVar()   , 'units': "Ohms"},
            'Ca'  : {'label': 'C[a]'       , 'var': tk.IntVar()   , 'units': "nFarads"},
            '_Cb' : {'label': 'C[b]'       , 'var': tk.IntVar()   , 'units': "MOhms"},
            '_Ra' : {'label': 'R[a]'       , 'var': tk.DoubleVar(), 'units': "Ohms"},
            '_Rb' : {'label': 'R[b]'       , 'var': tk.DoubleVar(), 'units': "Ohms"},
            '_Rc' : {'label': 'R[damp]'    , 'var': tk.DoubleVar(), 'units': "Ohms"},
            '_Ca' : {'label': 'C[a]'       , 'var': tk.DoubleVar(), 'units': "nFarads"},
            'G'   : {'label': 'G'          , 'var': tk.DoubleVar(), 'units': 'Reduction Factor'},
            'Rm'  : {'label': 'Rm'         , 'var': tk.IntVar()   , 'units': "Effective Damping Resistance"},
            'F'   : {'label': 'C[a]'       , 'var': tk.DoubleVar(), 'units': 'LPF -3dB'},
            '_G'  : {'label': 'G'          , 'var': tk.DoubleVar(), 'units': 'Reduction Factor'},
            '_Rm' : {'label': 'Rm'         , 'var': tk.IntVar()   , 'units': "Effective Damping Resistance"},
            '_F'  : {'label': 'C[a]'       , 'var': tk.DoubleVar(), 'units': 'LPF -3dB'},
            '_G_tol'  : {'label': ''       , 'var': tk.StringVar(), 'units': '', 'threshold': 0.02},
            '_Rm_tol' : {'label': ''       , 'var': tk.StringVar(), 'units': '', 'threshold': 0.02},
            '_F_tol'  : {'label': ''       , 'var': tk.StringVar(), 'units': '', 'threshold': 0.05},
        }
        
        for name in ['_Ra', '_Rb', '_Rc', '_Ca']:
            payload[name]['var'].trace("w", self.recalculate_callback)
        
        return payload
    
    
    def setup_components(self, master):
        
        left = tk.Frame(master, bg="#aaccbb")
        left.pack(side=tk.LEFT, expand=False, fill=tk.BOTH)
        right = tk.Frame(master, bg="#aaaabb")
        right.pack(side=tk.LEFT, expand=True, fill=tk.BOTH)
        
        self.input_frame = self.setup_input_frame(left)
        self.output_frame = self.setup_output_frame(right, have_gain_reduction=True)
        # only max motion before clipping and sensitivity 
        # are displayed on startup
        self.create_input_fields(self.input_frame, ['X', 'S'])
        
        
    def setup_input_frame(self, master):
    
        tk.Label(master, text="Geophone Input Parameters", bg=master['bg'], 
                 font=self.frame_title_font).pack(side=tk.TOP, fill=tk.Y, pady=10)
        
        frame = tk.Frame(master, bg=master['bg'])
        frame.pack(expand=False, fill=tk.Y)
        
        info_frame = tk.Frame(master, bg=master['bg'])
        info_frame.pack(side=tk.TOP, fill=tk.Y, pady=10)
        
        info_data = [
            '* We recommend 26 mm/s to match the Nanometrics Trillium Compact',
            '** This LPF is followed by a multiple pole digital LPF with a -3 dB frequency at ~40 Hz' 
        ]
        for info_text in info_data:
            tk.Label(info_frame, text=info_text, bg=info_frame['bg'], wraplength=400, 
                justify=tk.LEFT, font=self.small_font).pack(side=tk.TOP, fill=tk.Y, anchor=tk.W)

        tk.Label(master, text=__version__, bg=master['bg'], justify=tk.LEFT, 
            font=self.medium_font).pack(side=tk.BOTTOM, fill=tk.Y, anchor=tk.W)
            
        tk.Button(master, text="Calculate", 
            command=self.calculate_callback).pack(side=tk.BOTTOM, fill=tk.Y, pady=20)
          
        # frame content changes depending on gain (G) value 
        return frame
    
    
    def setup_output_frame(self, master, have_gain_reduction=True):
    
        for child in master.winfo_children():
            child.destroy()
        
        top = tk.Frame(master)
        top.pack(side=tk.TOP, expand=True, fill=tk.BOTH)
        
        bottom = tk.Frame(master, bg=master['bg'])
        bottom.pack(side=tk.TOP, expand=True, fill=tk.BOTH, pady=(0, 20))
        
        if have_gain_reduction:
            left = tk.Frame(top, bg=master['bg'])
            left.pack(side=tk.LEFT, expand=True, fill=tk.BOTH)
            
            right = tk.Frame(top, bg=master['bg'])
            right.pack(side=tk.LEFT, expand=True, fill=tk.BOTH)
            
            self.setup_circuit_values(left)
            self.setup_components_selection(right, entry_state='normal')
            self.setup_error_table(bottom)
        else:
            self.setup_components_selection(top, entry_state='readonly')
            self.setup_max_reading_frame(bottom)
            
        return master
        
        
    def create_input_fields(self, frame, fieldnames):
    
        # easier to fill the frame from scratch
        # rather than keeping track of visible/hidden elements
        for child in frame.winfo_children():
            child.destroy()
    
        # display elements
        for row, name in enumerate(fieldnames):
            tk.Label(frame, text=self.inputs[name]['label'] + ":", bg=frame['bg']).grid(row=row, 
                column=0, sticky="E", padx=1, pady=1)
            tk.Entry(frame, textvariable=self.inputs[name]['var'], width=8).grid(row=row, 
                column=1, sticky="NSEW", padx=1, pady=1)
            # the 3rd column can be either a dropdown or a plain label
            if name in self.multi_units:
                var, unit_values, conv = self.multi_units[name]
                optmenu = tk.OptionMenu(frame, var, *unit_values)
                optmenu.config(font=self.medium_font, width=6)
                optmenu.grid(row=row, column=2, sticky="W", padx=1, pady=1)
            else: 
                tk.Label(frame, text=self.inputs[name]['units'], bg=frame['bg']).grid(row=row, 
                    column=2, sticky="W", padx=1, pady=1)
        
        # set equal row sizes (optmenu breaks uniform heights)
        for row in range(frame.grid_size()[1]):
            frame.grid_rowconfigure(row, weight=1, uniform="uniform")
        
        
    def setup_circuit_values(self, master):
        tk.Label(master, text="Circuit Values", bg=master['bg'], 
                 font=self.frame_title_font).pack(side=tk.TOP, fill=tk.Y, pady=10)
                 
        frame = tk.Frame(master, bg=master['bg'])
        frame.pack(expand=False, fill=tk.Y)
        
        tk.Label(frame, text="Theoretical Values", font=self.table_title_font, 
            bg=frame['bg']).grid(row=0, column=0, sticky="NSEW", columnspan=3, padx=1, pady=(1, 5))
        
        # a dummy emty label (cannot get a row offset with grid)
        tk.Label(frame, text="", font=self.table_title_font, 
            bg=frame['bg']).grid(row=1, column=0, sticky="NSEW", columnspan=3, padx=1, pady=(1, 5))
        
        output_names = ['Rb', 'Rc', 'Ca']
        for row, name in enumerate(output_names, start=2):
            tk.Label(frame, text=self.outputs[name]['label'] + ":", bg=frame['bg']).grid(row=row, 
                column=0, sticky="E", padx=1, pady=1)
            tk.Entry(frame, textvariable=self.outputs[name]['var'], 
                state='readonly', width=8).grid(row=row, 
                column=1, sticky="NSEW", padx=1, pady=1)
            tk.Label(frame, text=self.outputs[name]['units'], bg=frame['bg']).grid(row=row, 
                column=2, sticky="W", padx=2, pady=1)
        
        
    def setup_components_selection(self, master, entry_state='normal'):

        tk.Label(master, text="Standard Component Selections", bg=master['bg'], 
                 font=self.frame_title_font).pack(side=tk.TOP, fill=tk.Y, pady=10)
                 
        frame = tk.Frame(master, bg=master['bg'])
        frame.pack(expand=False, fill=tk.Y)
        
        tk.Label(frame, text="Chosen (Actual) Circuit Values", font=self.table_title_font, 
            bg=frame['bg']).grid(row=0, column=0, sticky="NSEW", columnspan=3, padx=1, pady=(1, 5))
        
        output_names = ['_Ra', '_Rb', '_Rc', '_Ca', '_Cb']
        for row, name in enumerate(output_names, start=1):
            # _Cb never changes, so it is always in readonly state
            state = 'readonly' if name == '_Cb' else entry_state
            tk.Label(frame, text=self.outputs[name]['label'] + ":", bg=frame['bg']).grid(row=row, 
                column=0, sticky="E", padx=1, pady=1)
            tk.Entry(frame, textvariable=self.outputs[name]['var'], width=6, state=state).grid(row=row, 
                column=1, sticky="NSEW", padx=1, pady=1)
            tk.Label(frame, text=self.outputs[name]['units'], bg=frame['bg']).grid(row=row, 
                column=2, sticky="W", padx=2, pady=1)
            
            
    def setup_max_reading_frame(self, master):
        X = self.inputs['X']['var'].get()
        S = self.inputs['S']['var'].get()
        reading = int(max_reading(S, X, scale=2**23))
        fraction = max_reading(S, X, scale=1.0)
        
        text = "Maximum Digitizer Reading: {}".format(reading)
        tk.Label(master, text=text, bg=master['bg']).pack(side=tk.TOP, 
            fill=tk.Y, anchor=tk.W, pady=(10, 0))
        text = "which is {:.1f}% of max. possible full scale.".format(100 * fraction)
        tk.Label(master, text=text, bg=master['bg']).pack(side=tk.TOP, 
            fill=tk.Y, anchor=tk.W, pady=(0, 10))
        
        text = "R[Damp] is either not installed or factory specified."
        tk.Label(master, text=text, bg=master['bg'], 
            font=self.small_font).pack(side=tk.BOTTOM, fill=tk.Y, anchor=tk.W, pady=5)
        
        
    def setup_error_table(self, master):
        tk.Label(master, text="Error Table", bg=master['bg'], 
                 font=self.frame_title_font).pack(side=tk.TOP, fill=tk.Y, pady=10)
        
        frame = tk.Frame(master, bg=master['bg'])
        frame.pack(expand=False, fill=tk.Y)
        
        header = ['Value', 'Theoretical', 'Actual', 'Error', 'Description']
        for column, text in enumerate(header):
             tk.Label(frame, text=text, font=self.table_title_font, 
                bg=frame['bg']).grid(row=0, column=column, sticky="NSEW", padx=10, pady=1)
        
        self.color_entries = {}   # entries changing its bg color based on the value
        output_names = ['G', 'Rm', 'F']
        for row, name in enumerate(output_names, start=1):
            tk.Label(frame, text=self.outputs[name]['label'], bg=frame['bg']).grid(row=row, 
                column=0, sticky="E", padx=1, pady=1)
            tk.Entry(frame, textvariable=self.outputs[name]['var'], width=8, state='readonly',
                justify='center').grid(row=row, column=1, sticky="NSEW", padx=1, pady=1)
            tk.Entry(frame, textvariable=self.outputs["_" + name]['var'], width=8, state='readonly',
                justify='center').grid(row=row, column=2, sticky="NSEW", padx=1, pady=1)
            tol_name = "_" + name + "_tol"
            entry = tk.Entry(frame, textvariable=self.outputs[tol_name]['var'], width=8, 
                state='readonly', justify='center')
            entry.grid(row=row, column=3, sticky="NSEW", padx=1, pady=1)
            tk.Label(frame, text=self.outputs[name]['units'], bg=frame['bg']).grid(row=row, 
                column=4, sticky="W", padx=1, pady=1)
                
            self.color_entries[tol_name] = entry
    
    
    def pull_inputs(self):
        # can raise ValueError
        inputs = {name: self.inputs[name]['var'].get() for name in self.inputs}
            
        # convert values with dropdown units to uniform format
        for key in self.multi_units:
            variable, units, conv_map = self.multi_units[key]
            scaling_func = conv_map[variable.get()]
            inputs[key] = scaling_func(inputs[key])
            
        return inputs
        
        
    def push_outputs(self, outputs):
    
        # populate output fields
        for key, value in outputs.items():
            if '_tol' in key: # percentage
                formatted = "{:.2f}%".format(value * 100)
            elif np.isnan(value):
                formatted = value
            elif value < 1:
                formatted = round(value, 4)
            elif isinstance(self.outputs[key]['var'], tk.IntVar):
                formatted = int(round(value))
            elif round(value) == int(value):
                formatted = int(round(value))
            else:
                formatted = round(value, 1)
            self.outputs[key]['var'].set(formatted)

        # set bg color for entries in error column
        for name, entry in self.color_entries.items():
            if name in outputs:
                color = '#dd2222' if outputs[name] > self.outputs[name]['threshold'] else '#22dd22'
                entry.config(readonlybackground=color)
        
    
    def calculate_callback(self, show_additional = False):
        
        try: # handle non-numeric input
            inputs = self.pull_inputs()
        except ValueError:
            tkMessageBox.showerror("Error", "One or more input fields is empty")
            return
        G = gain_reduction(inputs['S'], inputs['X'])
        # a case when additional input is needed
        if (G < 1.0 and not self.have_all_inputs) or show_additional:
            self.have_all_inputs = True
            input_names = ['X', 'S', 'Rs', 'Rdr', 'D1', 'D2', 'F']
            self.create_input_fields(self.input_frame, input_names)
            self.output_frame.pack_forget()
            return
        
        if G < 1.0:
            outputs = self.calculate_gain_reduction(inputs)
        else:
            self.have_all_inputs = False
            input_names = ['X', 'S']
            self.create_input_fields(self.input_frame, input_names)
            outputs = self.calculate_nogain()
        
        if outputs:  # otherwise input validation failed
            self.output_frame = self.setup_output_frame(self.output_frame, 
                have_gain_reduction=(G < 1.0))
            self.push_outputs(outputs)
            self.output_frame.pack(side=tk.LEFT, expand=True, fill=tk.BOTH)
    
    def check_inputs_correct(self, data):
	if data['D1'] < 0 or data['D1'] > 1:
	    tkMessageBox.showerror("Error", "{} is an incorrect value. Must be less than 1.0".format(data['D1']))
	    return False
	return True;

    def calculate_gain_reduction(self, inputs, selected_values=None):
        # validate inputs
        if any(item <= 0 for item in inputs.values()):
            tkMessageBox.showerror("Error", "Only positive values expected")
            return {}
        
        # do some math
	if not self.check_inputs_correct(inputs):
	    return
        dsol = theoretical_solutions(inputs)
        if not check_solutions(dsol):
            tkMessageBox.showerror("Error", "Invalid solutions encountered")
            return {}
        
        outputs = {}
        
        if selected_values is None:
            components = select_component_values(dsol)
            outputs.update(components)
        else:  # use provided component values rather than selecting from the table
            components = selected_values
        
        Rb = components['_Ra'] + components['_Rb']
        errors = compute_errors(dsol, Rb, components['_Rc'], components['_Ca'])
        outputs.update(errors)
        
        return outputs
        
    
    def calculate_nogain(self):
     
        outputs = {
            '_Ra': 3300,
            '_Rb': 0,
            '_Rc': np.nan,
            '_Ca': 100,  # nFarads
            '_Cb': 1
        }
        return outputs
        
        
    def recalculate_callback(self, *args):
    
        # prevent calculations until all inputs are present
        if not self.have_all_inputs:
            return
            
        try: # handle non-numeric input
            inputs = self.pull_inputs()
        except ValueError:
            tkMessageBox.showerror("Error", "One or more input fields is empty")
            return
        
        values = {
            name: self.outputs[name]['var'].get() 
            for name in ['_Ra', '_Rb', '_Rc', '_Ca']
        }
        
        if any(item < 0 for item in values.values()) or values['_Ca'] == 0:
            # don't show an error since callback responds to changes dynamically
            return
        
        outputs = self.calculate_gain_reduction(inputs, selected_values=values)
        self.push_outputs(outputs)
        
        
    def run(self):
        self.root.update()
        size = self.root.winfo_width(), self.root.winfo_height()
        # prevent from resizing past minimal dimensions
        self.root.minsize(*size)
        # hide the output frame on startup
        self.output_frame.pack_forget()
        self.root.mainloop()
        


if __name__ == '__main__':

    app = Application()
    app.run()

